export const root = "bk-root"
export const slider_title = "bk-slider-title"
export const slider_value = "bk-slider-value"
export default `.bk-root .bk-slider-title{white-space:nowrap;}.bk-root .bk-slider-value{font-weight:600;}`
